import datetime
import getpass
import json
import logging
import os
import time

from kensu.client import *
from kensu.utils.simple_cache import *
from kensu.utils.dsl import mapping_strategies
from kensu.utils.dsl.extractors import Extractors
from kensu.utils.dsl.lineage_builder import LineageBuilder
from kensu.utils.helpers import to_hash_key
from kensu.utils.injection import Injection
from kensu.pandas import DataFrame,Series


class Kensu(object):
    # list of triples i, o, mapping strategy
    # i and o are either one or a list of triples (object, DS, SC)
    dependencies = []
    dependencies_mapping = []
    dependencies_per_columns = {}
    real_schema_df = {}
    sent_runs = []
    data_collectors = {}
    model={}


    UNKNOWN_PHYSICAL_LOCATION = PhysicalLocation(name="Unknown", lat=0.12341234, lon=0.12341234,
                                                 pk=PhysicalLocationPK(city="Unknown", country="Unknown"))

    @staticmethod
    def discover_user_name():
        return getpass.getuser()

    @staticmethod
    def get_git_repo():
        cur_dir = os.path.realpath(os.path.curdir)
        try:
            import git
            try:
                git_repo = git.Repo(cur_dir, search_parent_directories=True)
                return git_repo
            except git.GitError as e:
                logging.warn("kensu-py was unable to identify a git repo. The working dir is not a git repo?")
                pass
        except ImportError as e:
            logging.warn("Install GitPython for a maximum context about the GIT code repo if any")
            pass

    @staticmethod
    def discover_code_location():
        cur_dir = os.path.realpath(os.path.curdir)
        code_location = cur_dir
        git_repo = Kensu.get_git_repo()
        if git_repo is not None:
            remote = git_repo.remote()
            code_location = next(remote.urls, code_location)
        return code_location

    @staticmethod
    def discover_code_version():

        code_version = datetime.datetime.now().isoformat()

        git_repo = Kensu.get_git_repo()
        if git_repo is not None:
            code_version = git_repo.head.commit.hexsha
            if git_repo.is_dirty():
                code_version = code_version + " (dirty)"
        return code_version

    def __init__(self, api_url=None, auth_token=None, process_name=None,
                 user_name=None, code_location=None, init_context=True, do_report=True, report_to_file=False, offline_file_name=None, reporter=None, **kwargs):
        """
        """ 
        kensu_host = self.get_kensu_host(api_url)
        kensu_auth_token = auth_token

        self.cache = SimpleCache()

        self.extractors = Extractors()
        pandas_support = kwargs["pandas_support"] if "pandas_support" in kwargs else True
        sklearn_support = kwargs["sklearn_support"] if "sklearn_support" in kwargs else True
        bigquery_support = kwargs["bigquery_support"] if "bigquery_support" in kwargs else True
        tensorflow_support = kwargs["tensorflow_support"] if "tensorflow_support" in kwargs else True
        project_names = kwargs["project_names"] if "project_names" in kwargs else []
        environment = kwargs["environment"] if "environment" in kwargs else None
        timestamp = kwargs["timestamp"] if "timestamp" in kwargs else None
        logical_naming = kwargs["logical_naming"] if "logical_naming" in kwargs else None
        mapping = kwargs["mapping"] if "mapping" in kwargs else None
        report_in_mem = kwargs["report_in_mem"] if "report_in_mem" in kwargs else False
        if "get_code_version" in kwargs and kwargs["get_code_version"] is not None:
            get_code_version = kwargs["get_code_version"]
        else:
            get_code_version = Kensu.discover_code_version

        self.extractors.add_default_supports(pandas_support=pandas_support, sklearn_support=sklearn_support,bigquery_support=bigquery_support,tensorflow_support=tensorflow_support)

        self.kensu_api = KensuEntitiesApi()
        self.kensu_api.api_client.host = kensu_host
        self.kensu_api.api_client.default_headers["X-Auth-Token"] = kensu_auth_token

        if timestamp is not None:
            self.kensu_api.api_client.default_headers["X-Entity-Creation-Time"] = timestamp
        else:
            timestamp = datetime.datetime.now().timestamp()*1000

        # add function to Kensu entities
        injection = Injection()
        injection.set_reporter(reporter)
        injection.set_do_report(do_report, offline_file_name=offline_file_name, report_to_file=report_to_file)
        injection.set_kensu_api(self.kensu_api)

        self.user = None
        self.code_base = None
        self.code_version = None
        self.timestamp = timestamp
        self.logical_naming = logical_naming
        self.mapping = mapping
        self.report_in_mem = report_in_mem

        self.set_default_physical_location(Kensu.UNKNOWN_PHYSICAL_LOCATION)
        # can be updated using set_default_physical_location
        self.init_context(process_name=process_name, user_name=user_name, code_location=code_location,
                          get_code_version=get_code_version, project_names=project_names, environment=environment, timestamp=timestamp)



    # sets the api url using host if passed, otherwise gets KENSU_API_URL
    def get_kensu_host(self, host=None):
        if host is None:
            if "KENSU_API_URL" in os.environ:
                kensu_host = os.environ["KENSU_API_URL"]
            else:
                kensu_host = None
        else:
            kensu_host = host

        return kensu_host


    def init_context(self, process_name=None, user_name=None, code_location=None, get_code_version=None, project_names=None,environment=None,timestamp=None):

        if user_name is None:
            user_name = Kensu.discover_user_name()

        if code_location is None:
            code_location = Kensu.discover_code_location()

        self.user = User(pk=UserPK(user_name))._report()
        self.code_base = CodeBase(pk=CodeBasePK(code_location))._report()
        self.timestamp = timestamp
        if get_code_version is None:
            if timestamp is not None: # this is weird though...
                version = datetime.datetime.fromtimestamp(timestamp/1000).isoformat()
            else:
                version = Kensu.discover_code_version()
        else:
            version = get_code_version()
        self.code_version = CodeVersion(maintainers_refs=[self.user.to_ref()],
                                        pk=CodeVersionPK(version=version,
                                                         codebase_ref=self.code_base.to_ref()))._report()
        if process_name is None:
            if "__file__" in globals():
                process_name = os.path.basename(os.path.realpath(__file__))
            else:
                raise Exception("Can't determine `process_name`, maybe is this running from a Notebook?")
        self.process = Process(pk=ProcessPK(qualified_name=process_name))._report()

        
        if project_names is None:
            self.project_refs = []

        else:
            self.project_refs = [Project(pk=ProjectPK(name=n))._report().to_ref() for n in project_names]

        #TODO timestamp for historical data

        process_run_name = process_name + "@" + datetime.datetime.now().isoformat()
        self.process_run = ProcessRun(
            pk=ProcessRunPK(process_ref=self.process.to_ref(), qualified_name=process_run_name)
            , launched_by_user_ref=self.user.to_ref()
            , executed_code_version_ref=self.code_version.to_ref()
            , projects_refs=self.project_refs
            , environment = environment
        )._report()
        


    def set_default_physical_location(self, pl):
        self.default_physical_location = pl
        pl._report()
        self.default_physical_location_ref = pl.to_ref()

    def get_dependencies(self):
        return self.dependencies

    def get_dependencies_mapping(self):
        return self.dependencies_mapping

    def add_dependencies_mapping(self, guid, col, from_guid, from_col, type):
        dep = {'GUID': guid,
               'COLUMNS': col,
               'FROM_ID': from_guid,
               'FROM_COLUMNS': from_col,
               'TYPE': type}
        self.dependencies_mapping.append(dep)

    def add_dependency(self, i, o, mapping_strategy=mapping_strategies.FULL):
        if not isinstance(i, tuple):
            (ids, isc) = self.extractors.extract_data_source_and_schema(i, self.default_physical_location_ref)
            i = (i, ids, isc)

        if not isinstance(o, tuple):
            (ods, osc) = self.extractors.extract_data_source_and_schema(o, self.default_physical_location_ref)
            o = (o, ods, osc)

        self.dependencies.append((i, o, mapping_strategy))

    def add_dependencies(self, ins, outs, mapping_strategy=mapping_strategies.FULL):
        new_ins = []
        for i in ins:
            if not isinstance(i, tuple):
                (ids, isc) = self.extractors.extract_data_source_and_schema(i, self.default_physical_location_ref)
                i = (i, ids, isc)
                new_ins.append(i)

        new_outs = []
        for o in outs:
            if not isinstance(o, tuple):
                (ods, osc) = self.extractors.extract_data_source_and_schema(o, self.default_physical_location_ref)
                o = (o, ods, osc)
                new_outs.append(o)

        self.dependencies.append((new_ins, new_outs, mapping_strategy))

    def in_mem(self, var_name):
        return "in-memory-data://" + self.process.pk.qualified_name + "/" + var_name

    @property
    def s(self):
        return self.start_lineage(True)

    def start_lineage(self, report_stats=True):
        lineage_builder = LineageBuilder(self, report_stats)
        return lineage_builder

    # if the new_lineage has a model training in it (output),
    #   then kwargs will be pass to the function to compute metrics
    #     ex: kwargs["y_test"] can refer to the test set to compute CV metrics
    def new_lineage(self, process_lineage_dependencies, report_stats=True, **kwargs):
        data_flow = [d.toSchemaLineageDependencyDef() for d in process_lineage_dependencies]

        inputs = ",".join(sorted([d.from_schema_ref.by_guid for d in data_flow]))
        outputs = ",".join(sorted([d.to_schema_ref.by_guid for d in data_flow]))

        lineage = ProcessLineage(name=inputs+"->"+outputs,
                                 operation_logic="APPEND",
                                 # FIXME? => add control and the function level like report_stats
                                 pk=ProcessLineagePK(
                                     process_ref=self.process.to_ref(),
                                     data_flow=data_flow
                                 )
                                 )._report()

        if self.timestamp is None:
            self.timestamp=int(time.time()) * 1000

        lineage_run = LineageRun(pk=LineageRunPK(
            lineage_ref=lineage.to_ref(),
            process_run_ref=self.process_run.to_ref(),
            timestamp=self.timestamp
        )
        )._report()

        data_flow_inputs = list(
            {to_hash_key(d.input_schema): (d.input_schema, d.input) for d in process_lineage_dependencies}.values())
        data_flow_outputs = list(
            {to_hash_key(d.output_schema): (d.output_schema, d.output) for d in process_lineage_dependencies}.values())

        for (schema, df) in (data_flow_inputs + data_flow_outputs):
            stats = self.extractors.extract_stats(df)
            if report_stats and stats is not None:
                DataStats(pk=DataStatsPK(schema_ref=schema.to_ref(),
                                         lineage_run_ref=lineage_run.to_ref()),
                          stats=stats,
                          extra_as_json=None)._report()

        # TODO Machine Learning part for OUTPUTS ONLY (right ?)
        for (schema, df) in data_flow_outputs:
            ml = self.extractors.extract_machine_learning_info(df)
            if ml is not None:
                model = Model(ModelPK(ml["name"]))._report()
                model_training = ModelTraining(ModelTrainingPK(model_ref=model.to_ref(),
                                                               process_lineage_ref=lineage.to_ref())
                                               )._report()

                metrics = self.extractors.extract_machine_learning_metrics(df, **kwargs)
                if len(metrics) > 0:
                    hp = self.extractors.extract_machine_learning_hyper_parameters(df)
                    ModelMetrics(pk=ModelMetricsPK(
                        model_training_ref=model_training.to_ref(),
                        lineage_run_ref=lineage_run.to_ref(),
                        stored_in_schema_ref=schema.to_ref()
                    ),
                        metrics=metrics,
                        hyper_params_as_json=json.dumps(hp)
                    )._report()

    def report_with_mapping(self):
        import pandas as pd
        deps = self.dependencies_mapping
        ddf = pd.DataFrame(deps)
        df = ddf.set_index(['GUID', 'COLUMNS', 'FROM_ID']).groupby(['GUID', 'COLUMNS', 'FROM_ID']).agg(list)
        df = df[~df.index.duplicated(keep='first')].reset_index()

        unique_ids = list(df['GUID'].unique())

        if self.report_in_mem:
            for element in unique_ids:
                dataflow = []
                dependencies = df[df['GUID'] == element]
                data = {}
                for row in dependencies.iterrows():
                    info = row[1]
                    from_columns = [str(x) for x in info['FROM_COLUMNS']]
                    data[str(info['COLUMNS'])] = from_columns
                schema_dep = SchemaLineageDependencyDef(from_schema_ref=SchemaRef(by_guid=info['FROM_ID']),
                                                        to_schema_ref=SchemaRef(by_guid=info['GUID']),
                                                        column_data_dependencies=data)
                dataflow.append(schema_dep)

                lineage = ProcessLineage(name='Lineage', operation_logic='APPEND',
                                         pk=ProcessLineagePK(process_ref=ProcessRef(by_guid=self.process.to_guid()),
                                                             data_flow=dataflow))._report()

                if lineage.to_guid() not in self.sent_runs:
                    lineage_run = LineageRun(pk=LineageRunPK(lineage_ref=ProcessLineageRef(by_guid=lineage.to_guid()),
                                                             process_run_ref=ProcessRunRef(
                                                                 by_guid=self.process_run.to_guid()),
                                                             timestamp=round(self.timestamp)))._report()
                    self.sent_runs.append(lineage.to_guid())

        else:

            dependencies_per_columns = {}
            for element in unique_ids:
                if element in self.real_schema_df:
                    sub_df = df[df['GUID'] == element]
                    for row in sub_df.iterrows():
                        info = row[1]
                        destination_guid = info['GUID']
                        guid = info['GUID']
                        origin_column = info['COLUMNS']
                        column = info['COLUMNS']
                        all_deps = df
                        self.create_dependencies(destination_guid, guid, origin_column, column, all_deps,
                                            dependencies_per_columns)

            dataflows = {}
            for destination_guid in dependencies_per_columns:
                if dependencies_per_columns[destination_guid] != {}:
                    dataflows[destination_guid] = {}
                    for column in dependencies_per_columns[destination_guid]:
                        for origin_guid in dependencies_per_columns[destination_guid][column]:
                            if origin_guid not in dataflows[destination_guid] and origin_guid!=destination_guid:
                                dataflows[destination_guid][origin_guid] = {}
                                dataflows[destination_guid][origin_guid][column]=list(set(dependencies_per_columns[destination_guid][column][origin_guid]))
                            elif origin_guid!=destination_guid:
                                dataflows[destination_guid][origin_guid][column] = \
                                list(set(dependencies_per_columns[destination_guid][column][origin_guid]))

            for to_guid in dataflows:
                schemas_pk = set()
                from_pks = set()
                dataflow = []
                is_ml_model = False
                for from_guid in dataflows[to_guid]:

                    schema_dep = SchemaLineageDependencyDef(from_schema_ref=SchemaRef(by_guid=from_guid),
                                                            to_schema_ref=SchemaRef(by_guid=to_guid),
                                                            column_data_dependencies=dataflows[to_guid][from_guid])
                    if to_guid in self.model:
                        is_ml_model = True
                    dataflow.append(schema_dep)
                    schemas_pk.add(from_guid)
                    from_pks.add(from_guid)
                    schemas_pk.add(to_guid)


                lineage = ProcessLineage(name='Lineage to %s from %s' %(to_guid, (',').join(list(from_pks))), operation_logic='APPEND',
                                         pk=ProcessLineagePK(
                                             process_ref=ProcessRef(by_guid=self.process.to_guid()),
                                             data_flow=dataflow))._report()



                if lineage.to_guid() not in self.sent_runs:
                    lineage_run = LineageRun(
                        pk=LineageRunPK(lineage_ref=ProcessLineageRef(by_guid=lineage.to_guid()),
                                        process_run_ref=ProcessRunRef(by_guid=self.process_run.to_guid()),
                                        timestamp=round(self.timestamp)))._report()
                    self.sent_runs.append(lineage.to_guid())

                    for schema in schemas_pk:
                        stats_df = self.real_schema_df[schema]
                        try:
                            stats = self.extractors.extract_stats(stats_df)
                        except:
                            # FIXME weird... should be fine to delete (and try,except too)
                            if isinstance(stats_df, pd.DataFrame) or isinstance(stats_df, DataFrame) or isinstance(stats_df,Series) or isinstance(stats_df,pd.Series):
                                stats = self.extractors.extract_stats(stats_df)
                        if stats is not None:
                            DataStats(pk=DataStatsPK(schema_ref=SchemaRef(by_guid=schema),
                                                     lineage_run_ref=LineageRunRef(by_guid=lineage_run.to_guid())),
                                      stats=stats,
                                      extra_as_json=None)._report()
                        #FIXME should be using extractors instead
                        if is_ml_model:
                            model_name = self.model[to_guid][1]
                            metrics = self.model[to_guid][2]
                            import json
                            hyperparams = json.dumps(self.model[to_guid][3])

                            model = Model(ModelPK(name=model_name))._report()
                            train = ModelTraining(pk=ModelTrainingPK(model_ref=ModelRef(by_guid=model.to_guid()),
                                                                     process_lineage_ref=ProcessLineageRef(
                                                                         by_guid=lineage.to_guid())))._report()
                            r=ModelMetrics(pk=ModelMetricsPK(model_training_ref=ModelTrainingRef(by_guid=train.to_guid()),
                                                           lineage_run_ref=LineageRunRef(by_guid=lineage_run.to_guid()),
                                                           stored_in_schema_ref=SchemaRef(by_guid=to_guid)),
                                         metrics=metrics, hyper_params_as_json=hyperparams)._report()



    dependencies_per_columns = {}
    def create_dependencies(self,destination_guid, guid, origin_column, column, all_deps,
                            dependencies_per_columns_rt):
        self.dependencies_per_columns = dependencies_per_columns_rt
        filtered_dependencies = all_deps[all_deps['GUID'] == guid]

        filtered_dependencies = filtered_dependencies[filtered_dependencies['COLUMNS'] == column]
        if destination_guid in self.dependencies_per_columns:
            for row in filtered_dependencies.iterrows():
                row = row[1]

                if row['FROM_ID'] in self.real_schema_df:
                    if origin_column in self.dependencies_per_columns[destination_guid]:
                        if row['FROM_ID'] in self.dependencies_per_columns[destination_guid][origin_column]:
                            self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] = \
                            self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] + \
                            row['FROM_COLUMNS']
                        else:
                            self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] = row[
                                'FROM_COLUMNS']
                    else:
                        self.dependencies_per_columns[destination_guid][origin_column] = {}
                        self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] = row[
                            'FROM_COLUMNS']
                        # dependencies_per_columns[guid][row['FROM_ID']] = row['FROM_COLUMNS']
                else:
                    guid = row['FROM_ID']
                    columns = row['FROM_COLUMNS']
                    for column in columns:
                        self.create_dependencies(destination_guid, guid, origin_column, column, all_deps,
                                            self.dependencies_per_columns)
        else:
            self.dependencies_per_columns[destination_guid] = {}
            self.create_dependencies(destination_guid, guid, origin_column, column, all_deps,
                                self.dependencies_per_columns)


import sklearn.metrics as mk
from kensu.numpy import ndarray
from kensu.pandas import DataFrame,Series

def wrap_classification(method):
    def wrapper(*args, **kwargs):

        new_args=[]
        for arg in args:
            if isinstance(arg,DataFrame):
                new_args.append(arg.get_df())
            elif isinstance(arg,Series):
                new_args.append(arg.get_s())
            elif isinstance(arg, ndarray):
                new_args.append(arg.get_nd())
            else:
                new_args.append(arg)

        result = method(*new_args, **kwargs)

        return result

    wrapper.__doc__ = method.__doc__
    return wrapper


classification_report = wrap_classification(mk.classification_report)



#  python -m unittest discover -s tests/unit
'''

import json
import unittest

import urllib3

from kensu.client import ApiClient
from kensu.client import rest
from kensu.utils.kensu import Kensu
from kensu.utils.kensu_class_handlers import KensuClassHandlers


class TestRawEvents(unittest.TestCase):
    TypeToClass = {
        "PHYSICAL_LOCATION": "PhysicalLocation"
        , "USER": "User"
        , "CODE_BASE": "CodeBase"
        , "CODE_VERSION": "CodeVersion"
        , "PROCESS": "Process"
        , "PROCESS_RUN": "ProcessRun"
        , "DATA_SOURCE": "DataSource"
        , "SCHEMA": "Schema"
        , "PROCESS_LINEAGE": "ProcessLineage"
        , "LINEAGE_RUN": "LineageRun"
        , "DATA_STATS": "DataStats"
        , "MODEL": "Model"
        , "MODEL_TRAINING": "ModelTraining"
        , "MODEL_METRICS": "ModelMetrics"
    }

    kensu = Kensu(api_url = "http://example.com", init_context = False)

    def setUp(self):
        self.data = {}
        # Load test data
        ac = ApiClient()
        with open('tests/unit/fixtures/raw-events.json') as fp:
            for line in fp:
                j = json.loads(line)
                if j["action"] == "add_entity": # only entities
                    jp = j["jsonPayload"]
                    r = rest.RESTResponse(urllib3.response.HTTPResponse(json.dumps(jp)))
                    e = ac.deserialize(r, TestRawEvents.TypeToClass[j["entity"]])
                    self.data[j["generatedEntityGUID"]] = e
                
    def test_guids(self):
        for [expected_guid, entity] in self.data.items() :
            guid = KensuClassHandlers.guid(entity)
            self.assertEqual(guid, expected_guid)
                
    def test_to_guids(self):
        for [expected_guid, entity] in self.data.items() :
            guid = entity.to_guid()
            self.assertEqual(guid, expected_guid)


if __name__ == '__main__':
    unittest.main()

'''
from kensu.client.models import *
from kensu.utils.injection import Injection
Injection().kensu_inject_entities('api')
v= DataSource(name='predict/jan-customers-data.csv',format='csv',pk=DataSourcePK(location='file:/Users/kensu/Customers/Kensu/oreilly/data/predict/jan-customers-data.csv',physical_location_ref=PhysicalLocationRef(by_guid='k-d2f40e99e5dd4c9fc9c634b15a7fb03073191c0158e52a572769df8c05f59b7b')))
print(v.to_guid())


from hashlib import sha256

import numpy

import kensu
from kensu.client import FieldDef, SchemaPK, Schema, DataSource, DataSourcePK
from kensu.utils.dsl.extractors import ExtractorSupport
from kensu.utils.helpers import singleton

@singleton
class ndarraySupport(ExtractorSupport):  # should extends some KensuSupport class

    def is_supporting(self, nd):
        return isinstance(nd, numpy.ndarray)

    def is_machine_learning(self, nd):
        return False

    def skip_wr(self, nd):
        if isinstance(nd, kensu.numpy.ndarray):
            return nd.get_nd()
        else:
            return nd

    # return list of FieldDef
    def extract_schema_fields(self, nd):
        nd = self.skip_wr(nd)
        return [FieldDef(name="value", field_type=str(nd.dtype), nullable=True)]

    def extract_location(self, nd, location):
        if location is not None:
            return location
        else:
            nd = self.skip_wr(nd)
            return "in-mem://AN_ID" + sha256(str(nd.tolist()).encode("utf-8")).hexdigest()

    def extract_format(self, nd, fmt):
        if fmt is not None:
            return fmt
        else:
            nd = self.skip_wr(nd)
            return nd.__class__.__name__

    def tk(self, k, k1): return k + '.' + k1

    # return dict of doubles (stats)
    def extract_stats(self, nd):
        nd = self.skip_wr(nd)
        #TODO use scipy.stats?
        #return {self.tk(k, k1): v for k, o in nd.describe().to_dict().items() for k1, v in o.items()}
        return {"count": len(nd)}

    def extract_data_source(self, nd, pl, **kwargs):
        nd = self.skip_wr(nd)
        location = self.extract_location(nd, kwargs.get("location"))
        fmt = self.extract_format(nd, kwargs.get("format"))

        if location is None or fmt is None:
            raise Exception(
                "cannot report new pandas dataframe without location ({}) a format provided ({})!".format(location, fmt))

        ds_pk = DataSourcePK(location=location, physical_location_ref=pl)
        name = location
        ds = DataSource(name=name, format=fmt, categories=[], pk=ds_pk)
        return ds

    def extract_schema(self, data_source, nd):
        nd = self.skip_wr(nd)
        fields = self.extract_schema_fields(nd)
        sc_pk = SchemaPK(data_source.to_ref(), fields=fields)
        schema = Schema(name="schema:"+data_source.name, pk=sc_pk)
        return schema

    def extract_data_source_and_schema(self, nd, pl, **kwargs):
        nd = self.skip_wr(nd)
        ds = self.extract_data_source(nd, pl, **kwargs)
        sc = self.extract_schema(ds, nd)
        return ds, sc



from hashlib import sha256

import pandas as pd
import numpy as np

import kensu
from kensu.client import *
from kensu.utils.dsl.extractors import ExtractorSupport
from kensu.utils.helpers import singleton

@singleton
class KensuPandasSupport(ExtractorSupport):  # should extends some KensuSupport class

    def is_supporting(self, df):
        return isinstance(df, pd.DataFrame) or isinstance(df, pd.Series)

    def is_machine_learning(self, df):
        return False

    def skip_wr(self, df):
        if isinstance(df, kensu.pandas.DataFrame):
            return df.get_df()
        elif isinstance(df, kensu.pandas.Series):
            return df.get_s()
        else:
            return df

    # return list of FieldDef
    def extract_schema_fields(self, df):
        df = self.skip_wr(df)
        if isinstance(df, pd.DataFrame):
            return [FieldDef(name=str(k), field_type=v.name, nullable=True) for (k, v) in df.dtypes.to_dict().items()]
        elif isinstance(df, pd.Series):
            return [FieldDef(name=df.name, field_type=str(df.dtypes), nullable=True)]

    def extract_location(self, df, location):
        
        if location is not None:
            return location
        else:
            df = self.skip_wr(df)
            return "in-mem://AN_ID" + sha256(str(df.to_dict()).encode("utf-8")).hexdigest() +'/in-mem-transformation'
            

    def extract_format(self, df, fmt):
        if fmt is not None:
            return fmt
        else:
            df = self.skip_wr(df)
            return df.__class__.__name__

    def tk(self, k, k1): return k + '.' + k1

    # return dict of doubles (stats)
    def extract_stats(self, df):
        df = self.skip_wr(df)
        if isinstance(df, pd.DataFrame):
            stats_dict = {self.tk(k, k1): v for k, o in df.describe().to_dict().items() for k1, v in o.items() if type(v) in [int,float,np.int64]}
            for key,item in stats_dict.copy().items():
                if type(item) == np.int64:
                    stats_dict[key] = int(item)
                if np.isnan(item):
                    del stats_dict[key]
            return stats_dict
        elif isinstance(df, pd.Series):
            return {k: v for k, v in df.describe().to_dict().items() if type(v) in [int,float] }

    def extract_data_source(self, df, pl, **kwargs):

        logical_naming = kwargs["logical_naming"] if "logical_naming" in kwargs else None

        df = self.skip_wr(df)
        location = self.extract_location(df, kwargs.get("location"))
        fmt = self.extract_format(df, kwargs.get("format"))

        if location is None or fmt is None:
            raise Exception(
                "cannot report new pandas dataframe without location ({}) a format provided ({})!".format(location, fmt))

        ds_pk = DataSourcePK(location=location, physical_location_ref=pl)
        name = ('/').join(location.split('/')[-2:])
        if logical_naming == 'File':
            logical_category = location.split('/')[-1]
            ds = DataSource(name=name, format=fmt, categories=['logical::'+logical_category], pk=ds_pk)
        else:
            ds = DataSource(name=name, format=fmt, categories=[], pk=ds_pk)
        return ds

    def extract_schema(self, data_source, df):
        df = self.skip_wr(df)
        fields = self.extract_schema_fields(df)
        sc_pk = SchemaPK(data_source.to_ref(), fields=fields)
        schema = Schema(name="schema:"+data_source.name, pk=sc_pk)
        return schema

    def extract_data_source_and_schema(self, df, pl, **kwargs):
        df = self.skip_wr(df)
        ds = self.extract_data_source(df, pl, **kwargs)
        sc = self.extract_schema(ds, df)
        return ds, sc


#  python -m unittest discover -s tests/unit

import logging
import sys
import unittest

import kensu.pandas as pd
from kensu.client import ApiClient
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.dsl import mapping_strategies
from kensu.utils.dsl.mapping_strategies import Strategy

log_format = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=log_format)


class TestPandas(unittest.TestCase):
    token = "eyJhbGciOiJIUzI1NiJ9.eyIkaW50X3Blcm1zIjpbXSwic3ViIjoib3JnLnBhYzRqLmNvcmUucHJvZmlsZS5Db21tb25Qcm9maWxlI3NhbW15IiwidG9rZW5faWQiOiI4NmYxMWU0Ni02NDk1LTRkNWItODhjZS01ODk1NTAxNDM0NzUiLCJhcHBncm91cF9pZCI6IjUyZTVkNDNjLTM0ZGYtNGRkZi1hM2Y3LTIwODU2Y2MzOWY2ZCIsIiRpbnRfcm9sZXMiOlsiYXBwIl0sImV4cCI6MTg4ODE0MTk2NiwiaWF0IjoxNTcyNzgxOTY2fQ.4CdaN9a80xc1ry6aT52MFxc33vb0nJVXX5TdGzdoJ7E"
    kensu = KensuProvider().initKensu(api_url="https://api.sandbox.kensu.io", auth_token=token, init_context=True)

    # FIXME... dunno ... that sounds rather nasty
    # def _constructor(self):
    #     print("yolo")
    #     return pd.DataFrame
    # opd.DataFrame._constructor = property(_constructor)

    def setUp(self):
        self.ac = ApiClient()
        self.dataset = 'tests/unit/fixtures/Macroeconomical.csv'

    def test_one(self):
        dataframe = pd.read_csv(self.dataset, sep=';')

        t = 11

        # lineage to send to kensu
        newdata = dataframe.iloc[:, [0, 4, 5]]
        self.kensu.add_dependency(dataframe, newdata, mapping_strategy=mapping_strategies.DIRECT)
        newdata.to_json('output/pd_iloc_test_{}.json'.format(t))

        # other lineage to send to kensu
        # *WARN*;
        #  we need to do this because otherwise, the object is mutated and unicity is kept between newdata and newdata2
        newdata_kensu = self.kensu.extractors.extract_data_source_and_schema(newdata, self.kensu.default_physical_location_ref)
        print(newdata_kensu)
        newdata2 = newdata
        newdata2.loc[:, 'Total'] = newdata2.loc[:, 'Spirtis Price'] + newdata2.loc[:, 'Wine Price']
        self.kensu.add_dependency((newdata, newdata_kensu[0], newdata_kensu[1]), newdata2, mapping_strategy=mapping_strategies.DIRECT.or_else(
            Strategy.from_dict({"Total": ['Spirtis Price', 'Wine Price']})))
        newdata2.to_json('output/pd_loc_test_{}.json'.format(t))


if __name__ == '__main__':
    unittest.main()



from kensu.client import *
from kensu.utils.dsl.extractors import ExtractorSupport
from kensu.utils.helpers import singleton
from hashlib import sha256


@singleton
class KensuSKLearnSupport(ExtractorSupport):
    def is_supporting(self, m):
        return m.__class__.__module__.startswith("sklearn") or m.__class__.__module__.startswith("kensu.sklearn")

    def is_machine_learning(self, df):
        return True  # FIXME... this is quite abusive

    # return dict of doubles (stats)
    def extract_stats(self, m):
        # todo (ideas): hist coefs. Or batch several runs and keep stats per coeff
        return {}

    def extract_machine_learning_info(self, m):
        return {
            "name": m.__class__.__module__ + "." + m.__class__.__name__
        }

    # FIXME: this is only for LinearRegression...
    def extract_schema(self, data_source, model):
        fields = [
            FieldDef(name="intercept_", field_type="float", nullable=False),
            FieldDef(name="coef_", field_type="array<float>", nullable=False)
        ]
        sc_pk = SchemaPK(data_source.to_ref(), fields=fields)
        schema = Schema(name="schema:" + data_source.name, pk=sc_pk)
        return schema



    def extract_location(self, model, location):

        if location is not None:
            return location
        else:
            try:
                return "in-mem://model/" + sha256(str(model.coef_).encode("utf-8")).hexdigest() + '/in-mem-transformation'
            except:
                return "in-mem://model/" + sha256(
                    str(model.leaf_size).encode("utf-8")).hexdigest() + '/in-mem-transformation'



    def extract_data_source(self, model, pl, **kwargs):
        logical_naming = kwargs["logical_naming"] if "logical_naming" in kwargs else None

        location = self.extract_location(model, kwargs.get("location"))
        fmt = kwargs["format"] if 'format' in kwargs else 'Model'

        if location is None or fmt is None:
            raise Exception(
                "cannot report new model dataframe without location ({}) a format provided ({})!".format(location,
                                                                                                          fmt))

        ds_pk = DataSourcePK(location=location, physical_location_ref=pl)
        name = ('/').join(location.split('/')[-3:])
        if logical_naming == 'File':
            logical_category = location.split('/')[-1]
            ds = DataSource(name=name, format=fmt, categories=['logical::' + logical_category], pk=ds_pk)
        else:
            ds = DataSource(name=name, format=fmt, categories=[], pk=ds_pk)
        return ds


    # FIXME
    #  THIS can be extended, refactored, ...
    def extract_machine_learning_metrics(self, m, **kwargs):
        from sklearn import metrics
        from kensu.sklearn.metrics import classification_report

        def compute_metrics(X_name,y_name,prefix='train.'):
            ml_metrics = {}
            if 'classification' in kwargs:
                classification = kwargs['classification']
            else:
                classification = False

            if y_name in kwargs and X_name in kwargs:
                x = kwargs[X_name]
                y = kwargs[y_name]

                if classification:
                    import pandas
                    pred_y = m.predict(x)
                    e=classification_report(y,pred_y,output_dict=True)
                    df = pandas.json_normalize(e, sep='.')
                    ml_metrics = (df.to_dict(orient='records')[0])
                    return ml_metrics

                else:
                    # FIXME... what is m has no predict function
                    pred_y = m.predict(x).get_nd()
                    # This pretty much for regression
                    # regression : http://scikit-learn.org/stable/modules/model_evaluation.html
                    try :
                        ml_metrics = {
                        #prefix+"score": m.score(x, y),
                        prefix+"explained_variance": metrics.explained_variance_score(pred_y, y),
                        prefix+"neg_mean_absolute_error": metrics.mean_absolute_error(pred_y, y),
                        prefix+"neg_mean_squared_error": metrics.mean_squared_error(pred_y, y),
                        prefix+"neg_mean_squared_log_error": metrics.mean_squared_log_error(pred_y, y),
                        prefix+"neg_median_absolute_error": metrics.median_absolute_error(pred_y, y),
                        prefix+"r2": metrics.r2_score(pred_y, y)
                        }
                    except :
                        ml_metrics = None
                    return ml_metrics

        m1= compute_metrics('x_test','y_test',prefix='test')
        m2= compute_metrics('x_train', 'y_train')
        m_total = {}

        if bool(m1):
            m_total.update(m1)
        if bool(m2):
            m_total.update(m2)
        return m_total


    def extract_machine_learning_hyper_parameters(self, m):
        return m.get_params()



import re
from hashlib import sha1




def to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def singleton(cls, *args, **kw):
    instances = {}

    def _singleton():
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton


def to_hash_key(o):
    return sha1(str(o).encode()).hexdigest()


def eventually_report_in_mem(o):
    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()
    if kensu.report_in_mem:
        o._report()
    return o


def get_absolute_path(path):
    import os
    for prefix in ["abfs", "/abfs", "dbfs", "/dbfs"]:
        if path.startswith(prefix):
            path = path.replace(prefix, '')
            prefix = prefix.replace('/', '')
            return prefix + ':' + path
    return 'file:' + str(os.path.abspath(path))


def maybe_report(o, report):
    from kensu.utils.kensu_provider import KensuProvider
    dam = KensuProvider().instance()
    if report:
        o._report()
    return o


def extract_ksu_ds_schema(kensu, orig_variable, report=False, register_orig_data=False):
    ds = kensu.extractors.extract_data_source(orig_variable, kensu.default_physical_location_ref, logical_naming=kensu.logical_naming)
    schema = kensu.extractors.extract_schema(ds, orig_variable)
    maybe_report(ds, report=report)
    maybe_report(schema, report=report)
    if register_orig_data:
        kensu.real_schema_df[schema.to_guid()] = orig_variable
    return ds, schema



from hashlib import sha256

import pandas as pd
import google
from google.cloud.bigquery import Table

import kensu
from kensu.client import *
from kensu.google.cloud.bigquery import Client
from kensu.utils.dsl.extractors import ExtractorSupport
from kensu.utils.helpers import singleton
from kensu.utils.kensu_provider import KensuProvider


@singleton
class KensuBigQuerySupport(ExtractorSupport):  # should extends some KensuSupport class

    def is_supporting(self, table):
        return isinstance(table, google.cloud.bigquery.table.Table) or isinstance(df, google.cloud.bigquery.table.RowIterator)
    def is_machine_learning(self, df):
        return False

    def skip_wr(self, df):
        return df

    # return list of FieldDef
    def extract_schema_fields(self, df):
        if isinstance(df, google.cloud.bigquery.table.Table) or isinstance(df, google.cloud.bigquery.table.RowIterator):
            return [FieldDef(name=str(k.name), field_type=k.field_type, nullable=k.is_nullable) for k in df.schema]


    def extract_location(self, df, location):
        # FIXME => what to do for RowIterator?
        if isinstance(df, google.cloud.bigquery.table.Table):
            return "bigquery:/" + df.path
        else:
            return location


    def extract_format(self, df, fmt):
        # FIXME => should do something about RowIterator...
        return "BigQuery Table"

    def tk(self, k, k1): return k + '.' + k1

    def extract_table_stats(self, table: Table):
        r = {
            "nrows": table.num_rows,
        }
        kensu = KensuProvider().instance()
        client: Client = kensu.data_collectors['BigQuery']
        stats_aggs = {}
        for f in table.schema:
            # f.field_type is
            # ["STRING", "BYTES",
            # "INTEGER", "INT64",
            # "FLOAT", "FLOAT64",
            # "BOOLEAN", "BOOL",
            # "TIMESTAMP", "DATE", "TIME", "DATETIME", "GEOGRAPHY", "NUMERIC", "BIGNUMERIC",
            # "RECORD", "STRUCT",]
            if f.field_type in ["INTEGER", "INT", "FLOAT", "FLOAT64", "NUMERIC", "BIGNUMERIC"]:
                stats_aggs[f.name] = {"min": f"min({f.name})",
                                      "max": f"max({f.name})",
                                      "mean": f"avg({f.name})",
                                      "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
            elif f.field_type in ["BOOLEAN", "BOOL"]:
                stats_aggs[f.name] = {"true": f"sum(case {f.name} when true then 1 else 0 end)",
                                      "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
        selector = ",".join([v+" "+c+"_"+s for c, vs in stats_aggs.items() for s, v in vs.items()])
        sts = client.query(f"select {selector} from `{str(table.reference)}`")
        sts_result = sts.result()
        stats = None
        for row in sts_result:
            stats = row # there is only one anyway
        for k, vs in stats_aggs.items():
            for s in vs.keys():
                r[k + "." + s] = stats[k + "_" + s]
        return r

    # return dict of doubles (stats)
    def extract_stats(self, df):
        # df = self.skip_wr(df)
        if isinstance(df, google.cloud.bigquery.table.Table):
            return self.extract_table_stats(df)
        elif isinstance(df, google.cloud.bigquery.table.RowIterator):
            # FIXME -> TODO
            return None

    def extract_data_source(self, df, pl, **kwargs):
        logical_naming = kwargs["logical_naming"] if "logical_naming" in kwargs else None

        location = self.extract_location(df, kwargs.get("location"))
        fmt = self.extract_format(df, kwargs.get("format"))

        if location is None or fmt is None:
            raise Exception(
                "cannot report new pandas dataframe without location ({}) a format provided ({})!".format(location, fmt))

        ds_pk = DataSourcePK(location=location, physical_location_ref=pl)
        name = ('/').join(location.split('/')[-2:])
        if logical_naming == 'File':
            logical_category = location.split('/')[-1]
            ds = DataSource(name=name, format=fmt, categories=['logical::'+logical_category], pk=ds_pk)
        else:
            ds = DataSource(name=name, format=fmt, categories=[], pk=ds_pk)
        return ds

    def extract_schema(self, data_source, df):
        fields = self.extract_schema_fields(df)
        sc_pk = SchemaPK(data_source.to_ref(), fields=fields)
        schema = Schema(name="schema:"+data_source.name, pk=sc_pk)
        return schema

    def extract_data_source_and_schema(self, df, pl, **kwargs):
        ds = self.extract_data_source(df, pl, **kwargs)
        sc = self.extract_schema(ds, df)
        return ds, sc


from kensu.utils.helpers import singleton


class ExtractorSupport(object):
    def is_supporting(self, value):
        pass

    def extract_data_source(self, value, physical_location, **kwargs):
        pass

    def extract_schema(self, data_source, value):
        pass

    def extract_data_source_and_schema(self, value, physical_location):
        pass

    def extract_stats(self, value):
        pass

    def is_machine_learning(self, ml):
        pass

    def extract_machine_learning_info(self, ml):
        pass

    def extract_machine_learning_metrics(self, ml, **kwargs):
        pass

    def extract_machine_learning_hyper_parameters(self, ml):
        pass


@singleton
class Extractors(object):

    def __init__(self, ):
        self.supports = []

    def add_default_supports(self, pandas_support=True, sklearn_support=True, numpy_support=True, tensorflow_support=False, bigquery_support=False, generic_datasource_info_support=True):
        if pandas_support:
            from kensu.pandas.extractor import KensuPandasSupport
            self.add_support(KensuPandasSupport())

        if sklearn_support:
            from kensu.sklearn.extractor import KensuSKLearnSupport
            self.add_support(KensuSKLearnSupport())

        if numpy_support:
            from kensu.numpy.extractor import ndarraySupport
            self.add_support(ndarraySupport())

        if bigquery_support:
            from kensu.google.cloud.bigquery.extractor import KensuBigQuerySupport
            self.add_support(KensuBigQuerySupport())

        if generic_datasource_info_support:
            from kensu.utils.dsl.extractors.generic_datasource_info_support import GenericDatasourceInfoSupport
            self.add_support(GenericDatasourceInfoSupport())



    def add_support(self, support):
        self.supports.append(support)

    def extract_data_source(self, value, physical_location, **kwargs):
        for support in self.supports:
            if support.is_supporting(value):
                return support.extract_data_source(value, physical_location, **kwargs)

        raise Exception("Not supported object: " + str(value.__class__))

    def extract_schema(self, data_source, value):
        for support in self.supports:
            if support.is_supporting(value):
                return support.extract_schema(data_source, value)

        raise Exception("Not supported object: " + value.__class__)

    def extract_data_source_and_schema(self, value, physical_location):
        for support in self.supports:
            if support.is_supporting(value):
                return support.extract_data_source_and_schema(value, physical_location)

        raise Exception("Not supported object: " + value.__class__)

    def extract_stats(self, value):
        for support in self.supports:
            if support.is_supporting(value):
                return support.extract_stats(value)

        raise Exception("Not supported container: " + value.__class__)

    def extract_machine_learning_info(self, ml):
        for support in self.supports:
            if support.is_machine_learning(ml):
                if support.is_supporting(ml):
                    return support.extract_machine_learning_info(ml)
        # Not sure though... maybe there is no support for the model yet... how to figure out
        return None

    def extract_machine_learning_metrics(self, ml, **kwargs):
        for support in self.supports:
            if support.is_machine_learning(ml):
                if support.is_supporting(ml):
                    return support.extract_machine_learning_metrics(ml, **kwargs)

        raise Exception("Not supported container: " + ml.__class__)

    def extract_machine_learning_hyper_parameters(self, ml):
        for support in self.supports:
            if support.is_machine_learning(ml):
                if support.is_supporting(ml):
                    return support.extract_machine_learning_hyper_parameters(ml)

        raise Exception("Not supported container: " + ml.__class__)



import logging

import numpy as np

from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.dsl import mapping_strategies
from kensu.utils.helpers import eventually_report_in_mem, get_absolute_path

class ndarrayDelegator(object):
    SKIP_KENSU_FIELDS = ["_ndarray__k_nd", "INTERCEPTORS"]
    SKIP_KENSU_METHODS = ["get_nd", "kensu_init", "to_string"]

    def __getattribute__(self, name):
        attr_value = object.__getattribute__(self, name)
        if name == "_ndarray__k_nd":
            return attr_value
        elif name == "__class__":
            return attr_value
        elif hasattr(attr_value, '__call__'):
            # return ndarrayDelegator.handle_callable(self, name, attr_value)
            return attr_value
        else:
            return ndarrayDelegator.handle_field(self, name, attr_value)

    @staticmethod
    def wrap_returned_df(kensu_s, returned, name, original):
        if isinstance(returned, np.ndarray) \
                and not isinstance(returned, ndarray) \
                and original is not returned:
            logging.debug("Wrapping attr access resulting ndarray in KensuDF for property name=" + name)
            tmp = ndarray.using(returned)
            returned = tmp
        return returned

    @staticmethod
    def handle_field(kensu_df, name, attr_value):
        # _DataFrame__k_df => points to DataFrame.__k_df
        if name in ndarrayDelegator.SKIP_KENSU_FIELDS:
            # return Kensu DF attr value right away
            logging.debug("Kensu attribute name=" + name)
            return attr_value
        else:
            logging.debug("NOT Kensu, let's delegate attribute name=" + name)

            nd = kensu_df.get_nd()
            nd_attr = object.__getattribute__(nd, name)
            result = ndarrayDelegator.wrap_returned_df(kensu_df, nd_attr, name, nd)

            # adding the datasources, schemas and lineage to the dependencies
            if result is not None and isinstance(result, ndarray):
                kensu = KensuProvider().instance()
                orig_ds = kensu.extractors.extract_data_source(nd, kensu.default_physical_location_ref)._report()
                orig_sc = kensu.extractors.extract_schema(orig_ds, nd)._report()

                result_ds = kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref)._report()
                result_sc = kensu.extractors.extract_schema(result_ds, result)._report()

                kensu.add_dependency((nd, orig_ds, orig_sc), (result, result_ds, result_sc),
                                   mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            return result


class ndarray(ndarrayDelegator, np.ndarray):
    __k_nd = None

    # https://numpy.org/doc/stable/user/basics.subclassing.html
    def __new__(subtype, shape, dtype=float, buffer=None, offset=0, strides=None, order=None):
        obj = super(ndarray, subtype).__new__(subtype, shape, dtype, buffer, offset, strides, order)
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.__k_nd = getattr(obj, '__k_nd', None)

    @staticmethod
    def using(o):
        if isinstance(o, ndarray):
            d = ndarray.using(o.get_nd())
            return d
        else:
            d = ndarray(o.shape,o.dtype)
            d.kensu_init(o)
            return d

    def get_nd(self):
        # if the current Kensu ndarray isn't delegating... TODO.. need to have a type for this instead...
        if self.__k_nd is None:
            return self
        else:
            return self.__k_nd

    def kensu_init(self, d):
        self.__k_nd = d


def wrap_save(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()
        result = method(*args, **kwargs)


        loc = args[0]
        df = args[1]


        if df.__class__ == ndarray:

            location = get_absolute_path(loc)

            orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,
                                                         logical_naming=kensu.logical_naming))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, df))


            result_ds = kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref, location = location,
                                                           logical_naming=kensu.logical_naming)._report()
            result_sc = kensu.extractors.extract_schema(result_ds,df)._report()

            kensu.real_schema_df[result_sc.to_guid()] = result
            

            if kensu.mapping == True:
                for col in [s.name for s in result_sc.pk.fields]:
                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col),
                                                 "save")
                kensu.report_with_mapping()


        return result

    wrapper.__doc__ = method.__doc__
    return wrapper

savetext = wrap_save(np.savetxt)


def wrap_where(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()

        origins = []
        new_args = []
        condition = [args[0]]
        for item in args:
            from kensu.pandas import Series
            if isinstance(item,Series):
                new_args.append(item.get_s())
            else:
                new_args.append(item)
        for item in args[1:]:
            if isinstance(item,Series):
                origins.append(item)
        new_args = tuple(new_args)

        result = method(*new_args, **kwargs)

        result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref,
                                                       logical_naming=kensu.logical_naming))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

        for element in origins:
            df = element.get_s()

            orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,
                                                                                  logical_naming=kensu.logical_naming))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, df))

            if kensu.mapping == True:
                for col in [s.name for s in result_sc.pk.fields] :
                    for col_orig in [s.name for s in orig_sc.pk.fields]:

                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col_orig),
                                                     "Numpy Where")
        result = ndarray.using(result)
        return result

    wrapper.__doc__ = method.__doc__
    return wrapper


where = wrap_where(np.where)



from google.cloud.bigquery import client
from .job.query import QueryJob
from google.cloud.bigquery.retry import DEFAULT_RETRY
from kensu.utils.kensu_provider import KensuProvider

class Client(client.Client):


    def query(
        self,
        query,
        job_config=None,
        job_id=None,
        job_id_prefix=None,
        location=None,
        project=None,
        retry=DEFAULT_RETRY,
        timeout=None
    ):
        gclient = super(Client, self)
        j = gclient.query(query,
        job_config,
        job_id,
        job_id_prefix,
        location,
        project,
        retry,
        timeout)
        kensu = KensuProvider().instance()
        kensu.data_collectors['BigQuery'] = gclient

        return QueryJob.patch(j)


import pandas as pd

from .data_frame import DataFrame
from .data_frame import Series
from .data_frame import wrap_pandas_reader, wrap_pandas_get_dummies, wrap_merge
from .extractor import KensuPandasSupport

if hasattr(pd, "read_clipboard"):
    read_clipboard = wrap_pandas_reader(pd.read_clipboard)
if hasattr(pd, "read_csv"):
    read_csv = wrap_pandas_reader(pd.read_csv)
if hasattr(pd, "read_excel"):
    read_excel = wrap_pandas_reader(pd.read_excel)
if hasattr(pd, "read_feather"):
    read_feather = wrap_pandas_reader(pd.read_feather)
if hasattr(pd, "read_fwf"):
    read_fwf = wrap_pandas_reader(pd.read_fwf)
if hasattr(pd, "read_gbq"):
    read_gbq = wrap_pandas_reader(pd.read_gbq)
if hasattr(pd, "read_hdf"):
    read_hdf = wrap_pandas_reader(pd.read_hdf)
if hasattr(pd, "read_html"):
    read_html = wrap_pandas_reader(pd.read_html)
if hasattr(pd, "read_json"):
    read_json = wrap_pandas_reader(pd.read_json)
if hasattr(pd, "read_parquet"):
    read_parquet = wrap_pandas_reader(pd.read_parquet)
if hasattr(pd, "read_pickle"):
    read_pickle = wrap_pandas_reader(pd.read_pickle)
if hasattr(pd, "read_sas"):
    read_sas = wrap_pandas_reader(pd.read_sas)
if hasattr(pd, "read_sql"):
    read_sql = wrap_pandas_reader(pd.read_sql)
if hasattr(pd, "read_sql_query"):
    read_sql_query = wrap_pandas_reader(pd.read_sql_query)
if hasattr(pd, "read_sql_table"):
    read_sql_table = wrap_pandas_reader(pd.read_sql_table)
if hasattr(pd, "get_dummies"):
    get_dummies = wrap_pandas_get_dummies(pd.get_dummies)
if hasattr(pd, "merge"):
    merge = wrap_merge(pd.merge)




import logging

from functools import reduce
import re

import sklearn.preprocessing as pp

from kensu.numpy import ndarray
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.dsl import mapping_strategies

class StandardScalerDelegator(object):
    SKIP_KENSU_FIELDS = ["_StandardScaler__k_sc", "INTERCEPTORS"]
    SKIP_KENSU_METHODS = ["get_sc", "kensu_init", "to_string"]

    def __getattribute__(self, name):
        attr_value = object.__getattribute__(self, name)
        if name == "_StandardScaler__k_sc":
            return attr_value
        elif name == "__class__":
            return attr_value
        elif hasattr(attr_value, '__call__'):
            return StandardScalerDelegator.handle_callable(self, name, attr_value)
        else:
            #StandardScalerDelegator.handle_field(self, name, attr_value)
            return attr_value

    @staticmethod
    def wrap_returned_df(kensu_sc, returned, name, original):
        if isinstance(returned, pp.StandardScaler) \
                and not isinstance(returned, StandardScaler) \
                and original is not returned:
            logging.debug("Wrapping attr access resulting StandardScaler in KensuDF for property name=" + name)
            tmp = StandardScaler.using(returned)
            returned = tmp
        return returned

    @staticmethod
    def handle_callable(kensu_sc, name, attr_value):
        if name not in StandardScalerDelegator.SKIP_KENSU_METHODS:
            sc = object.__getattribute__(kensu_sc, "get_sc")()
            if isinstance(getattr(type(sc), name), property):
                delegated_pd_attr = object.__getattribute__(sc, name)
                # todo ? -> StandardScalerDelegator.handle_callable_property(kensu_sc, name, sc, delegated_pd_attr)
                return delegated_pd_attr
            else:
                delegated_pd_attr = object.__getattribute__(sc, name)
                docstring = delegated_pd_attr.__doc__
                return StandardScalerDelegator.create_function_wrapper(kensu_sc, name, docstring)
        else:
            logging.debug("Kensu function name=" + name)

            # return Kensu DF function result right away
            def w(*args, **kwargs):
                return attr_value(*args, **kwargs)
            return w

    @staticmethod
    def create_function_wrapper(kensu_sc, name, docstring):
        logging.debug("Kensu wrapping function name=" + name)

        def wrapper(*args, **kwargs):
            # call method on delegated df
            delegation_config = getattr(kensu_sc, 'INTERCEPTORS', None)

            for regex_attr_name in delegation_config:
                result_regex = regex_attr_name.search(name)
                if result_regex is not None:
                    # if we pass the regex, there is an interception defined, so we execute it
                    name_of_intercept_method = delegation_config[regex_attr_name]
                    intercept = object.__getattribute__(kensu_sc, name_of_intercept_method)
                    intercept(result_regex, *args, **kwargs)

            sc = kensu_sc.get_sc()
            s_attr = object.__getattribute__(sc, name)

            result = s_attr(*args, **kwargs)
            result = StandardScalerDelegator.wrap_returned_df(kensu_sc, result, name, sc)

            #adding the datasources, schemas and lineage to the dependencies
            if result_regex is None \
                    and result is not None:
                if name == "fit_transform":
                    kensu = KensuProvider().instance()
                    orig_o = args[0]
                    orig_ds = kensu.extractors.extract_data_source(orig_o, kensu.default_physical_location_ref)._report()
                    orig_sc = kensu.extractors.extract_schema(orig_ds, orig_o)._report()

                    result_ds = kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref)._report()
                    result_sc = kensu.extractors.extract_schema(result_ds, result)._report()

                    if not isinstance(result, ndarray):
                        result = ndarray.using(result)

                    kensu.add_dependency((orig_o, orig_ds, orig_sc), (result, result_ds, result_sc),
                                       mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            return result

        wrapper.__doc__ = docstring
        return wrapper

class StandardScaler(StandardScalerDelegator, pp.StandardScaler):
    INTERCEPTORS = {
        re.compile('to_(.*)'): 'k_to_format'
    }

    __k_sc = None

    @staticmethod
    def using(o):
        if isinstance(o, StandardScaler):
            d = StandardScaler.using(o.get_sc())
            return d
        else:
            d = StandardScaler()
            d.kensu_init(o)
            return d

    def get_sc(self):
        # if the current Kensu StandardScaler isn't delegating... TODO.. need to have a type for this instead...
        if self.__k_sc is None:
            return self
        else:
            return self.__k_sc

    def kensu_init(self, d):
        self.__k_sc = d



class KensuDatasourceAndSchema:
    def __init__(self,
                 ksu_ds,
                 ksu_schema,
                 f_get_stats=lambda: {}):
        """
        :param kensu.client.DataSource ksu_ds: Datasource info which was extracted from external system
        :param kensu.client.Schema ksu_schema: Schema info for the datasource
        :param () -> typing.Dict[string, float] f_get_stats: [Optional] a function to retrieve DataStats for this DataSource
        """
        self.ksu_ds = ksu_ds
        self.ksu_schema = ksu_schema
        # this allows expensive delaying computation of stats until they are needed (if at all)
        self.f_get_stats = f_get_stats

    @staticmethod
    def for_path_without_schema(ksu, ds_path, format=None, categories=None):
        from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
        pl_ref = ksu.UNKNOWN_PHYSICAL_LOCATION.to_ref()
        ds_pk = DataSourcePK(location=ds_path, physical_location_ref=pl_ref)
        ds = DataSource(name=ds_path, format=format, categories=categories, pk=ds_pk)
        fields = [
            FieldDef(name=str("unknown"), field_type="unknown", nullable=True),
        ]
        schema = Schema(name="schema:" + ds.name,
                        pk=SchemaPK(ds.to_ref(), fields=fields))
        return KensuDatasourceAndSchema(ksu_ds=ds, ksu_schema=schema)


class ExtDependencyEntry:
    def __init__(
            self,
            input_ds,  # type: KensuDatasourceAndSchema
            # {out_column: [input_column1, input_column2, ...]}
            lineage=None  # type: dict[str, list[str]]
    ):
        """
        :param KensuDatasourceAndSchema input_ds:
        :param typing.Dict[str, typing.List[str]]  lineage: format is  `{out_column: [input_column1, input_column2, ...]}`
        """
        self.input_ds = input_ds
        if lineage is None:
            lineage = {}
        self.lineage = lineage


class GenericComputedInMemDs:
    def __init__(
            self,
            inputs,  # type: list[KensuDatasourceAndSchema]
            lineage  # type: list[ExtDependencyEntry]
    ):
        self.lineage = lineage
        self.inputs = inputs

    def report(self, ksu, df_result, operation_type):
        from kensu.utils.helpers import extract_ksu_ds_schema
        from kensu.utils.dsl import mapping_strategies
        for input_ds in self.inputs:
            extract_ksu_ds_schema(ksu, input_ds, report=True, register_orig_data=True)
        # report output (if needed)
        result_ds, result_schema = extract_ksu_ds_schema(ksu, df_result, report=ksu.report_in_mem, register_orig_data=True)
        # register the lineage
        for dep in self.lineage:
            input_ds = dep.input_ds
            # at column level
            if ksu.mapping:
                # lineage = { out_column: [input_column1, ...] }
                for out_col, input_cols in dep.lineage.items():
                    for in_col in input_cols:
                        # FIXME: maybe there's an issue here is that result_schema DO NOT physically exist?
                        ksu.add_dependencies_mapping(guid=result_schema.to_guid(),
                                                     col=str(out_col),
                                                     from_guid=input_ds.ksu_schema.to_guid(),
                                                     from_col=str(in_col),
                                                     type=operation_type)
            # at datasource level (not column lineage info known)
            else:
                i=(input_ds, input_ds.ksu_ds, input_ds.ksu_schema)
                o=(df_result, result_ds, result_schema)
                ksu.add_dependency(i=i,
                                   o=o,
                                   mapping_strategy=mapping_strategies.FULL)


    @staticmethod
    def report_copy_without_schema(src, dest, operation_type):
        from kensu.utils.kensu_provider import KensuProvider
        ksu = KensuProvider().instance()
        input_ds = KensuDatasourceAndSchema.for_path_without_schema(ksu=ksu, ds_path=src)
        output_ds = KensuDatasourceAndSchema.for_path_without_schema(ksu=ksu, ds_path=dest)
        lineage_info = [ExtDependencyEntry(
            input_ds=input_ds,
            # FIXME: kensu-py lineage do not work without schema as well...
            lineage={
                "unknown": ["unknown"],
            })]
        inputs_lineage = GenericComputedInMemDs(inputs=[input_ds], lineage=lineage_info)
        # register lineage in KensuProvider
        inputs_lineage.report(ksu=ksu, df_result=output_ds, operation_type="sftp.put()")
        # actuly report the lineage and the write operation
        ksu.report_with_mapping()



#  python -m unittest discover -s tests/unit

import logging
import sys
import unittest

import kensu.pandas as pd
from kensu.client import ApiClient
from kensu.utils.kensu_provider import KensuProvider

log_format = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=log_format)

# Below is an example of how easily one could track a lineage of data conversion from an external library into pandas
# A real world example would be Apache Spark's DataFrame.toPandas() function,
# here to reduce need of heavy dependency we use a mocked version of such function


class FakeSparkDataFrame:

    def toPandas(self):
        import pandas as pd_orig
        data = [['tom', 10], ['nick', 15], ['juli', 14]]
        df = pd_orig.DataFrame(data, columns = ['out_field_name1', 'Age'])
        return df


def get_inputs_lineage_fn(ksu, df):
    from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema, GenericComputedInMemDs, \
        ExtDependencyEntry
    from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
    pl_ref = ksu.UNKNOWN_PHYSICAL_LOCATION.to_ref()
    ds_pk = DataSourcePK(location="hdfs://namenode/dataset1/partition1/file1.parquet", physical_location_ref=pl_ref)
    ds = DataSource(name="file1.parquet", format="parquet", categories=['logical::dataset1'], pk=ds_pk)
    fields = [
        FieldDef(name=str("in1"), field_type="string", nullable=True),
        FieldDef(name=str("in2"), field_type="string", nullable=True),
    ]
    schema = Schema(name="schema:" + ds.name,
                    pk=SchemaPK(ds.to_ref(), fields=fields))
    lineage_info = [ExtDependencyEntry(
        input_ds=KensuDatasourceAndSchema(ksu_ds=ds, ksu_schema=schema, f_get_stats=lambda: {'in1.max': 123.12}),
        lineage={
            "out_field_name1": ["in1", "in2"],
            "Age": ["in2"]

        })]
    return GenericComputedInMemDs(inputs=list([x.input_ds for x in lineage_info]), lineage=lineage_info)


def patch_external_transformation():
    print('patching FakeDataFrame.toPandas')
    FakeSparkDataFrame.toPandas = pd.data_frame.wrap_external_to_pandas_transformation(
        FakeSparkDataFrame.toPandas,
        get_inputs_lineage_fn
    )
    print('done patching FakeDataFrame.toPandas')


class TestExternalToPandas(unittest.TestCase):
    offline = True
    kensu = KensuProvider().initKensu(init_context=True,
                                      report_to_file=offline,
                                      offline_file_name='kensu-offline-to-pandas-test.jsonl',
                                      mapping=True, report_in_mem=True)

    def setUp(self):
        self.ac = ApiClient()
        patch_external_transformation()

    def test_one(self):
        FakeSparkDataFrame().\
            toPandas()\
            .to_csv('fake_spark_to_pandas_out')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2021 Kensu Inc
#
import logging

from kensu.pandas import DataFrame
from kensu.utils.kensu_provider import KensuProvider
import google.cloud.bigquery as bq
import google.cloud.bigquery.job as bqj
from google.cloud.bigquery import Table
import sqlparse

class QueryJob(bqj.QueryJob):

    @staticmethod
    def patch(job: bqj.QueryJob) -> bqj.QueryJob:
        return QueryJob.override_result(QueryJob.override_to_dataframe(job))


    @staticmethod
    def override_to_dataframe(job: bqj.QueryJob) -> bqj.QueryJob:
        f = job.to_dataframe

        def wrapper(*args, **kwargs):
            kensu = KensuProvider().instance()
            result = f(*args, **kwargs)
            df = DataFrame.using(result)

            read_ds = kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref,
                                                         logical_naming=kensu.logical_naming)._report()
            read_sc = kensu.extractors.extract_schema(read_ds, df)._report()

            query = job.query
            sq = sqlparse.parse(query)
            ids = list(filter(lambda x: isinstance(x, sqlparse.sql.Identifier), sq[0].tokens))
            client = kensu.data_collectors['BigQuery']
            for id in ids:
                name = (id.get_real_name()).strip('`')
                table = client.get_table(name)
                path = table.path
                location = "bigquery:/" + path
                fmt = "BigQuery Table"
                ds = kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref, location=location,
                                                        format=fmt, logical_naming=kensu.logical_naming)._report()
                sc = kensu.extractors.extract_schema(ds, table)._report()

                if kensu.mapping:
                    for col in result:
                        if col in [v.name for v in sc.pk.fields]:
                            dep = {'GUID': read_sc.to_guid(),
                                   'COLUMNS': col,
                                   'FROM_ID': sc.to_guid(),
                                   'FROM_COLUMNS': col,
                                   'TYPE': 'read'}
                            kensu.dependencies_mapping.append(dep)

                kensu.real_schema_df[sc.to_guid()] = result[[v.name for v in sc.pk.fields]]

            return df
        wrapper.__doc__ = f.__doc__
        setattr(job, 'to_dataframe', wrapper)
        return job

    @staticmethod
    def find_sql_identifiers(tokens):
        for t in tokens:
            if isinstance(t, sqlparse.sql.Identifier):
                if t.is_group and len(t.tokens) > 0:
                    # String values like "World" in `N == "World"` are also Identifier
                    # but their first child is of ttype `Token.Literal.String.Symbol`
                    # although table seems to have a first child of ttype `Token.Name`
                    if str(t.tokens[0].ttype) == "Token.Name":
                        # FIXME .. this is also returning the column names... (REF_GET_TABLE)
                        yield t
            elif t.is_group:
                yield from QueryJob.find_sql_identifiers(t)


    @staticmethod
    def table_to_kensu(table: Table):
        kensu = KensuProvider().instance()
        ds = kensu.extractors.extract_data_source(table, kensu.default_physical_location_ref,
                                                logical_naming=kensu.logical_naming)._report()
        sc = kensu.extractors.extract_schema(ds, table)._report()
        return ds, sc


    @staticmethod
    def get_table_info_for_id(client: bq.Client, id: sqlparse.sql.Identifier):
        try:
            name = (id.get_real_name()).strip('`')
            table = client.get_table(name)
            ds, sc = QueryJob.table_to_kensu(table)
            return table, ds, sc
        except:
            # FIXME this is because the current find_sql_identifiers also returns the column names...
            #  (see aboveREF_GET_TABLE)
            #  Therefore get_table of a column name should fail
            return None


    @staticmethod
    def get_table_infos_from_sql(client: bq.Client, query: str):
        sq = sqlparse.parse(query)
        ids = QueryJob.find_sql_identifiers(sq[0].tokens) # FIXME we only take the first element
        table_infos = list(filter(lambda x: x is not None, [QueryJob.get_table_info_for_id(client, id) for id in ids]))
        return table_infos


    @staticmethod
    def override_result(job: bqj.QueryJob) -> bqj.QueryJob:
        f = job.result
        def wrapper(*args, **kwargs):
            kensu = KensuProvider().instance()
            result = f(*args, **kwargs)
            client = kensu.data_collectors['BigQuery']
            # FIXME lots of copy paste from above function to_dataframe
            dest = job.destination
            if not dest:
                logging.debug("Not implemented job without destination")
            else:
                if isinstance(dest, bq.TableReference):
                    dest = client.get_table(dest)
                destination_ds, destination_sc = QueryJob.table_to_kensu(dest)
                kensu.real_schema_df[destination_sc.to_guid()] = dest
                dest_field_names = [f.name for f in destination_sc.pk.fields]
                table_infos = QueryJob.get_table_infos_from_sql(client, job.query)
                for table, ds, sc in table_infos:
                    if kensu.mapping:
                        sc_field_names = [v.name for v in sc.pk.fields]
                        for col in dest_field_names:
                            if col in sc_field_names:
                                dep = {'GUID': destination_sc.to_guid(),
                                       'COLUMNS': col,
                                       'FROM_ID': sc.to_guid(),
                                       'FROM_COLUMNS': col,
                                       'TYPE': 'read'}
                                kensu.dependencies_mapping.append(dep)

                    kensu.real_schema_df[sc.to_guid()] = table
                kensu.report_with_mapping()

            return result #TODO lineage and stuff if lost from here on

        wrapper.__doc__ = f.__doc__
        setattr(job, 'result', wrapper)
        return job




#! python -m py.test test_sftp.py

import logging
import sys
import os
import unittest
import pytest


from kensu import pysftp
from kensu.client import ApiClient
from kensu.utils.kensu_provider import KensuProvider

log_format = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=log_format)


@pytest.mark.usefixtures("sftpserver")
class TestSftp(unittest.TestCase):
    offline = True
    api_url = os.environ.get('KSU_API_URL') or ''
    auth_token = os.environ.get('KSU_API_TOKEN') or ''
    kensu = KensuProvider().initKensu(init_context=True,
                                      api_url=api_url,
                                      auth_token=auth_token,
                                      report_to_file=offline,
                                      project_names=['pysftp'],
                                      offline_file_name='kensu-offline-to-pysftp-test.jsonl',
                                      mapping=True, report_in_mem=True)

    def setUp(self):
        self.ac = ApiClient()

    # see https://docs.pytest.org/en/reorganize-docs/unittest.html#autouse-fixtures-and-accessing-other-fixtures
    @pytest.fixture(autouse=True)
    def init_sftpserver(self, sftpserver, tmpdir):
        self.sftpserver = sftpserver
        self.tmpfile = tmpdir.join("samplefile.txt")
        self.tmpfile.write("# testdata")

    def test_one(self):
        # uses https://pypi.org/project/pytest-sftpserver/
        with self.sftpserver.serve_content({'a_dir': {'somefile.txt': "File content"}}):
            cnopts = pysftp.CnOpts()
            cnopts.hostkeys = None
            with pysftp.Connection(host=self.sftpserver.host,
                                   port=self.sftpserver.port,
                                   username='user',
                                   password='pass',
                                   cnopts=cnopts) as sftp:
                print("Connection to sftp OK  ... ")
                local_file_path = self.tmpfile
                remote_file_path = f"/2021-05-01.csv"
                sftp.put(local_file_path, remote_file_path)
                directory_structure = sftp.listdir_attr()
                uploaded = False
                for attr in directory_structure:
                    print(attr.filename, attr)
                    if "/" + attr.filename == remote_file_path:
                        uploaded = True
                assert uploaded, True


if __name__ == '__main__':
    unittest.main()
